"""
    instagramy.plugins.manual_loading
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Parse data of Instagram with manual feeding of Instagram Data in offline mode.
    This classes are more useful for data analysis purpose. Store the data in json
    file and Parse this data by using this classes instead of sending multiple requests
    to the Instagram

    Usage Example
    -------------
    ::
        >>> import json
        >>> from instagramy import InstagramUser

        >>> user = InstagramUser('github')
        >>> user_data = user.user_data

        # store data of user in json file
        >>> with open('github_user.json', 'w') as file_obj:
        ...     json.dump(user_data, file_obj)

        >>> from instagramy.plugins.manual_loading import InstagramUser

        # using the stored data
        >>> with open('github_user.json', 'r') as file_obj:
        ...     user_data = json.load(file_obj)
        >>> user = InstagramUser(user_data)
        >>> user.number_of_followers

"""

from instagramy.core.parser import UserParser
from instagramy.core.parser import TagParser
from instagramy.core.parser import PostParser
from instagramy.core.parser import LocationParser


class InstagramUser(UserParser):
    r"""
    Parse the data of User from manual loading

    :param data: user_data from `instagramy.InstagramUser.user_data`
    """

    def __init__(self, data: dict):
        self.user_data = data


class InstagramPost(PostParser):
    r"""
    Parse the data of Post from manual loading

    :param data: post_data from `instagramy.InstagramPost.post_data`
    """

    def __init__(self, data: dict):
        self.post_data = data


class InstagramHashTag(TagParser):
    r"""
    Parse the data of hashtag from manual loading

    :param data: tag_data from `instagramy.InstagramHashTag.tag_data`
    """

    def __init__(self, data: dict):
        self.tag_data = data


class InstargramLocation(LocationParser):
    r"""
    Parse the data of location from manual loading

    :param data: location_data from `instagramy.InstagramLocation.location_data`
    """

    def __init__(self, data: dict):
        self.location_data = data



# Unable to fetch data from Instagram effectively
# So, Some data are already get stored in other website for testing the parsers
import unittest
import json

from instagramy.core.requests import get
from instagramy.plugins.manual_loading import *


# loading sample data for test from other website

user_data = json.loads(get("https://yogeshwaran01.herokuapp.com/user_data"))
post_data = json.loads(get("https://yogeshwaran01.herokuapp.com/post_data"))
tag_data = json.loads(get("https://yogeshwaran01.herokuapp.com/tag_data"))
location_data = json.loads(get("https://yogeshwaran01.herokuapp.com/location_data"))


class TestParsers(unittest.TestCase):
    """ Test case for all Instagramy Parsers """

    def test_InstagramUser(self):
        """ Test case for class `InstagramUser` """
        user = InstagramUser(user_data)
        self.assertEqual(user.biography, "Built for developers.")
        self.assertIsNone(user.connected_fb_page)
        self.assertFalse(user.followed_by_viewer)
        self.assertFalse(user.follows_viewer)
        self.assertEqual(user.fullname, "GitHub")
        self.assertFalse(user.has_blocked_viewer)
        self.assertFalse(user.has_country_block)
        self.assertFalse(user.has_blocked_viewer)
        self.assertFalse(user.is_blocked_by_viewer)
        self.assertFalse(user.is_joined_recently)
        self.assertFalse(user.is_private)
        self.assertTrue(user.is_verified)
        self.assertAlmostEqual(user.no_of_mutual_follower, 0)
        self.assertAlmostEqual(user.number_of_followers, 139340)
        self.assertAlmostEqual(user.number_of_followings, 20)
        self.assertAlmostEqual(user.number_of_posts, 182)
        self.assertEqual(user.username, "github")

    def test_InstagramPost(self):
        """ Test case for class `InstagramPost` """
        post = InstagramPost(post_data)
        self.assertEqual(post.author, "chilll_memes")
        self.assertEqual(
            post.caption,
            "Photo by CHILL MEMES in MRC Nagar with @memepattarai2.0. May be a meme of 3 people and text that says 'private Hospital nurse Chilll_ memes Govt Hospital nurse'.",
        )
        self.assertEqual(post.number_of_comments, 0)
        self.assertEqual(post.number_of_likes, 21)

    def test_InstagramHashtag(self):
        """ Test case for class `InstagramHashtag` """
        tag = InstagramHashTag(tag_data)
        self.assertEqual(tag.number_of_posts, 3600401)
        self.assertEqual(tag.tagname, "python")

    def test_InstagramLocation(self):
        """ Test case for class `InstagramLocation` """
        location = InstargramLocation(location_data)
        self.assertEqual(location.latitude, 32.86367)
        self.assertEqual(location.longitude, -117.212101)
        self.assertAlmostEqual(location.number_of_posts, 45580)



""" User Agents for webscraping """

# User-Agents for Web-Scraping
user_agents = [
    "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)",
    "Googlebot/2.1 (+http://www.googlebot.com/bot.html)",
    "Googlebot/2.1 (+http://www.google.com/bot.html)",
]



# -*- coding: utf-8 -*-
"""
    Instagramy
    ~~~~~~~~~~

    A python package for Instagram. It scarpe the Instagram contents.

    :license: MIT License
"""

__package__ = "instagramy"
__description__ = "Python Package for Instagram User, Posts, Hashtags and Locations data"
__url__ = "https://github.com/yogeshwaran01/instagramy"
__version__ = "4.4"
__author__ = "YOGESHWARAN R <yogeshin247@gmail.com>"
__license__ = "MIT License"
__copyright__ = "Copyright 2021 Yogeshwaran R"

__all__ = ["InstagramUser", "InstagramHashTag", "InstagramPost", "InstagramLocation"]

from .InstagramUser import InstagramUser
from .InstagramPost import InstagramPost
from .InstagramHashTag import InstagramHashTag
from .InstagramLocation import InstagramLocation



"""
    instagramy.InstagramLocation
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module scrape data
    of given Instagram Location.

    Usage Example
    -------------
    ::

        >>> from instagramy.InstagramLocation import InstagramLocation

        >>> location = InstagramLocation('977862530', 'mrc-nagar')
        >>> location.number_of_posts
        3119668
        >>> location.address

"""

from .core.parser import Parser
from .core.parser import Viewer
from .core.parser import LocationParser
from .core.exceptions import LocationNotFound
from .core.exceptions import RedirectionError
from .core.exceptions import HTTPError
from .core.cache import Cache
from .core.requests import get


class InstagramLocation(LocationParser):
    r"""
    Scrapes instagram location information
    `https://www.instagram.com/explore/locations/<location_id>/<slug>`
    `https://www.instagram.com/explore/locations/977862530/mrc-nagar`

    :param location_id: Location id of the location
    :param slug: slug name of the location
    :param sessionid (optional): Session id of Instagram which is in browser cookies
    :param from_cache (optional): Get data from the cache of instagramy not from instagram

    >>> location = InstagramLocation('977862530', 'mrc-nagar')
    >>> location.number_of_posts
    3119668
    >>> location.address
    """

    def __init__(self, location_id: str, slug: str, sessionid=None, from_cache=False):
        self.url = f"https://www.instagram.com/explore/locations/{location_id}/{slug}"
        self.sessionid = sessionid
        location = location_id + "_" + slug
        cache = Cache("location")
        if from_cache:
            if cache.is_exists(location):
                self.location_data = cache.read_cache(location)
            else:
                data = self.get_json()
                cache.make_cache(
                    location, data["entry_data"]["LocationsPage"][0]["graphql"]["location"]
                )
                self.location_data = data["entry_data"]["LocationsPage"][0]["graphql"]["location"]
        else:
            data = self.get_json()
            cache.make_cache(
                location, data["entry_data"]["LocationsPage"][0]["graphql"]["location"]
            )
            try:
                self.location_data = data["entry_data"]["LocationsPage"][0]["graphql"]["location"]
            except KeyError:
                raise RedirectionError
        if sessionid:
            try:
                self.viewer = Viewer(data=data["config"]["viewer"])
            except UnboundLocalError:
                self.viewer = None
        else:
            self.viewer = None

    def get_json(self) -> dict:
        """ Get Location information from Instagram """

        try:
            html = get(self.url, sessionid=self.sessionid)
        except HTTPError:
            raise LocationNotFound(self.url.split("/")[-2] + "_" + self.url.split("/")[-1])
        parser = Parser()
        parser.feed(html)
        return parser.Data

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.name}')"

    def __str__(self) -> str:
        return f"{self.name} has {self.number_of_posts} posts"



""" Parsers for Instagramy  """

import json
from datetime import datetime
from html.parser import HTMLParser
from collections import namedtuple

from .exceptions import RedirectionError
from .requests import get


def _nodes_classfier(nodes: list):
    post_lists = []
    for node in nodes:
        data = {}
        try:
            data["likes"] = node["node"]["edge_liked_by"]["count"]
        except (KeyError, TypeError):
            data["likes"] = None
        try:
            data["comments"] = node["node"]["edge_media_to_comment"]["count"]
        except (KeyError, TypeError):
            data["comments"] = None
        try:
            data["is_video"] = node["node"]["is_video"]
        except (KeyError, TypeError):
            data["is_video"] = None
        try:
            data["upload_time"] = datetime.fromtimestamp(
                node["node"]["taken_at_timestamp"]
            )
        except (KeyError, TypeError):
            data["upload_time"] = None
        try:
            data["caption"] = node["node"]["accessibility_caption"]
        except (KeyError, TypeError):
            data["caption"] = None
        try:
            data["shortcode"] = node["node"]["shortcode"]
        except (KeyError, TypeError):
            data["shortcode"] = None
        try:
            data["dimensions"] = node['node']["dimensions"]
        except (KeyError, IndexError):
            data["dimensions"] = None
        try:
            data[
                "post_url"
            ] = f'https://www.instagram.com/p/{node["node"]["shortcode"]}'
        except (KeyError, TypeError):
            data["post_url"] = None
        try:
            data["display_url"] = node["node"]["display_url"]
        except (KeyError, TypeError):
            data["display_url"] = None
        nt = namedtuple("Post", data.keys())(*data.values())
        post_lists.append(nt)

    return post_lists


class Parser(HTMLParser):

    """
    Class Parse the Static Html of the Instagram
    website and return the required Data as
    Python Dict

    This Class Inherits html.parser.HtmlParser
    """

    Data = {}

    def handle_data(self, data):
        if data.startswith("window._sharedData"):
            try:
                self.Data = json.loads(data[data.find('{"config"'): -1])
            except (KeyError, json.JSONDecodeError):
                raise RedirectionError
        else:
            pass


class Viewer:
    """
    User of Instagram currently Authenticated
    Parse the Current User data in Page
    """

    def __init__(self, **kwags):
        data = kwags.get("data")
        if data:
            self.user_data = data
        else:
            sessionid = kwags.get("sessionid")
            html = get("https://instagram.com", sessionid=sessionid)
            parser = Parser()
            parser.feed(html)
            self.user_data = parser.Data

    @property
    def username(self) -> str:
        """ Username of the given user """
        return self.user_data["username"]

    @property
    def fullname(self) -> str:
        """ Fullname of the given user """
        return self.user_data["full_name"]

    @property
    def biography(self) -> str:
        """ Biography of the given user """
        return self.user_data["biography"]

    @property
    def website(self) -> str:
        """ Website of the given user """
        return self.user_data["external_url"]

    @property
    def profile_picture_url(self) -> str:
        """ Profile picture url of the Given User """
        return self.user_data["profile_pic_url_hd"]

    @property
    def is_private(self) -> bool:
        """ Account type is Private """
        return self.user_data["is_private"]

    @property
    def is_joined_recently(self) -> bool:
        """ is user joined recently """
        return self.user_data["is_joined_recently"]

    @property
    def is_professional_account(self) -> bool:
        """ is user joined recently """
        return self.user_data["is_professional_account"]

    def __str__(self) -> str:
        return f"{self.fullname} ({self.username}) -> {self.biography}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.username}')"


class UserParser:
    """ Parse the required data of user store as property"""

    @property
    def username(self) -> str:
        """ Username of the given user """
        return self.user_data["username"]

    @property
    def fullname(self) -> str:
        """ Fullname of the given user """
        return self.user_data["full_name"]

    @property
    def biography(self) -> str:
        """ Biography of the given user """
        return self.user_data["biography"]

    @property
    def website(self) -> str:
        """ Website of the given user """
        return self.user_data["external_url"]

    @property
    def number_of_followers(self) -> int:
        """ No.of Followers of the given user """
        return self.user_data["edge_followed_by"]["count"]

    @property
    def number_of_followings(self) -> int:
        """ No.of Following of the given user """
        return self.user_data["edge_follow"]["count"]

    @property
    def number_of_posts(self) -> int:
        """ No.of Post of the given user """
        return self.user_data["edge_owner_to_timeline_media"]["count"]

    @property
    def profile_picture_url(self) -> str:
        """ Profile picture url of the Given User """
        return self.user_data["profile_pic_url_hd"]

    @property
    def is_verified(self) -> bool:
        """ Verification status of the user """
        return self.user_data["is_verified"]

    @property
    def is_private(self) -> bool:
        """ Account type is Private """
        return self.user_data["is_private"]

    @property
    def posts(self) -> list:
        """
        Top 12 posts data of the given user
        """

        posts_lists = []
        posts_details = self.user_data["edge_owner_to_timeline_media"]["edges"]
        for i in posts_details:
            data = {}
            try:
                data["likes"] = i["node"]["edge_liked_by"]["count"]
            except (KeyError, TypeError):
                data["likes"] = None
            try:
                data["comments"] = i["node"]["edge_media_to_comment"]["count"]
            except (KeyError, TypeError):
                data["comments"] = None
            try:
                data["caption"] = i["node"]["accessibility_caption"]
            except (KeyError, TypeError):
                data["caption"] = None
            try:
                data["is_video"] = i["node"]["is_video"]
            except (KeyError, TypeError):
                data["is_video"] = None
            try:
                data["timestamp"] = i["node"]["taken_at_timestamp"]
            except (KeyError, TypeError):
                data["timestamp"] = None
            try:
                data["location"] = i["node"]["location"]
            except (KeyError, TypeError):
                data["location"] = None
            try:
                data["shortcode"] = i["node"]["shortcode"]
            except (KeyError, TypeError):
                data["shortcode"] = None
            try:
                data[
                    "post_url"
                ] = f'https://www.instagram.com/p/{i["node"]["shortcode"]}/'
            except (KeyError, TypeError):
                data["post_url"] = None
            try:
                data["display_url"] = i["node"]["display_url"]
            except (KeyError, TypeError):
                data["display_url"] = None

            if i["node"]["is_video"]:
                data["video_url"] = i["node"]["video_url"]
                data["video_view_count"] = i["node"]["video_view_count"]
            if i["node"]["is_video"]:
                data["post_source"] = i["node"]["video_url"]
            else:
                data["post_source"] = i["node"]["display_url"]

            try:
                data["taken_at_timestamp"] = datetime.fromtimestamp(
                    i["node"]["taken_at_timestamp"]
                )
            except (KeyError, TypeError):
                data["taken_at_timestamp"] = None
            nt = namedtuple("Post", data.keys())(*data.values())
            posts_lists.append(nt)
        return posts_lists

    @property
    def posts_display_urls(self) -> list:
        """
        Top 12 posts picture url of the given user
        """

        return [i["display_url"] for i in self.posts]

    @property
    def is_joined_recently(self) -> bool:
        """ Is user joined recently """
        return self.user_data["is_joined_recently"]

    @property
    def other_info(self) -> dict:
        """
        Other information about user
        """
        return {
            "is_private": self.user_data["is_private"],
            "is_verified": self.user_data["is_verified"],
            "is_business_account": self.user_data["is_business_account"],
            "is_joined_recently": self.user_data["is_joined_recently"],
            "has_ar_effects": self.user_data["has_ar_effects"],
            "has_clips": self.user_data["has_clips"],
            "has_guides": self.user_data["has_guides"],
            "has_channel": self.user_data["has_channel"],
            "highlight_reel_count": self.user_data["highlight_reel_count"],
        }

    @property
    def follows_viewer(self) -> bool:
        """ Is user follows the Viewer """
        return self.user_data["follows_viewer"]

    @property
    def has_blocked_viewer(self) -> bool:
        """ Is user blocked the Viewer """
        return self.user_data["has_blocked_viewer"]

    @property
    def no_of_mutual_follower(self) -> bool:
        """ No of Mutual Followers """
        return self.user_data["edge_mutual_followed_by"]["count"]

    @property
    def requested_by_viewer(self) -> bool:
        """ Is viewer requested to follow user """
        return self.user_data["requested_by_viewer"]

    @property
    def is_blocked_by_viewer(self) -> bool:
        """ Is Viewer blocked the User """
        return self.user_data["blocked_by_viewer"]

    @property
    def restricted_by_viewer(self) -> bool:
        """ Is Viewer restricted the User """
        return self.user_data["restricted_by_viewer"]

    @property
    def has_country_block(self) -> bool:
        """ Is country blocked the User """
        return self.user_data["country_block"]

    @property
    def followed_by_viewer(self) -> bool:
        """ Is Viewer Follows the User """
        return self.user_data["followed_by_viewer"]

    @property
    def has_requested_viewer(self) -> bool:
        """ Is User requested the Viewer """
        return self.user_data["has_requested_viewer"]

    @property
    def connected_fb_page(self) -> bool:
        """ Connected Facebook page of User """
        return self.user_data["connected_fb_page"]


class PostParser:
    """ Parse the required data of post store as property"""

    @property
    def type_of_post(self) -> str:
        """ Type of the Post"""
        return self.post_data["__typename"]

    @property
    def display_url(self) -> str:
        """ Display url of the Image/Video """
        return self.post_data["display_url"]

    @property
    def upload_time(self) -> datetime:
        """ Upload Datetime of the Post """
        return datetime.fromtimestamp(self.post_data["taken_at_timestamp"])

    @property
    def number_of_likes(self) -> int:
        """ No.of Like is given post """
        return int(self.post_data["edge_media_preview_like"]["count"])

    @property
    def number_of_comments(self) -> int:
        """ No.of Comments is given post """
        return int(self.post_data["edge_media_to_parent_comment"]["count"])

    @property
    def author(self) -> str:
        """ Author of the Post """
        return self.post_data["owner"]["username"]

    @property
    def caption(self) -> str:
        """ Caption of the Post """
        return self.post_data["accessibility_caption"]

    @property
    def post_source(self) -> str:
        """ Post Image/Video Link """
        if self.post_data["is_video"]:
            return self.post_data["video_url"]
        return self.display_url

    @property
    def text(self) -> str:
        try:
            text = self.post_data["edge_media_to_caption"]["edges"][0]["node"]["text"]
            return text
        except (KeyError, IndexError):
            return None

    @property
    def location(self) -> str:
        location_data = self.post_data["location"]
        Location = namedtuple("Location", ["Address", "id", "slug", "name"])
        if location_data:
            return Location(
                location_data["address_json"],
                location_data["id"],
                location_data["slug"],
                location_data["name"],
            )
        return None


class TagParser:
    """ Parse the required data of tag store as property"""

    @property
    def tagname(self) -> str:
        """ Tagname of the Hagtag """
        return self.tag_data["name"]

    @property
    def profile_pic_url(self) -> str:
        """ Profile picture url of the Hagtag """
        return self.tag_data["profile_pic_url"]

    @property
    def number_of_posts(self) -> int:
        """ No.of posts in given Hashtag """
        return self.tag_data["edge_hashtag_to_media"]["count"]

    @property
    def top_posts(self) -> list:
        """
        Top post data (<70) in the given Hashtag
        """
        nodes = self.tag_data["edge_hashtag_to_media"]["edges"]
        return _nodes_classfier(nodes)

    @property
    def posts_display_urls(self) -> list:
        """
        Top post (<70) in the given Hashtag
        """
        return [i["display_url"] for i in self.top_posts]


class LocationParser:
    """ Parse the required data of location store as property"""

    @property
    def id(self) -> str:
        """ Location id of the location """
        return self.location_data["id"]

    @property
    def name(self) -> str:
        """ Name of the location """
        return self.location_data["name"]

    @property
    def latitude(self) -> int:
        """ Latitude of the location """
        return self.location_data["lat"]

    @property
    def longitude(self) -> int:
        """ Longitude of the location """
        return self.location_data["lng"]

    @property
    def slug(self) -> str:
        """ Slug of the location """
        return self.location_data["slug"]

    @property
    def website(self) -> str:
        """ Website of the location """
        return self.location_data["website"]

    @property
    def phone(self) -> str:
        """ Phone Number of the location """
        return self.location_data["phone"]

    @property
    def address(self) -> dict:
        """ Address of the location """
        return json.loads(self.location_data["address_json"])

    @property
    def profile_pic_url(self) -> dict:
        """ Profile Picture of the location """
        return self.location_data["profile_pic_url"]

    @property
    def number_of_posts(self) -> int:
        """ Number of post in the location """
        return self.location_data["edge_location_to_media"]["count"]

    @property
    def top_posts(self) -> int:
        """
        Top post data (<70) in the given Location
        """
        nodes_1 = self.location_data["edge_location_to_media"]["edges"]
        nodes_2 = self.location_data["edge_location_to_top_posts"]["edges"]

        return _nodes_classfier(nodes_1) + _nodes_classfier(nodes_2)



import sys
import argparse

from .InstagramUser import InstagramUser
from .InstagramPost import InstagramPost
from .InstagramHashTag import InstagramHashTag

from instagramy import __version__
from instagramy import __package__
from instagramy import __description__


def _pprint(data):
    for key, value in data.items():
        if value is None:
            value = ""
        print("{:<10} {:<10} ".format(key, value))


def _user(username):
    user = InstagramUser(username)
    return {
        "Username": user.username,
        "Name": user.fullname,
        "Biography": user.biography,
        "Followers": user.number_of_followers,
        "Following": user.number_of_followings,
        "Posts": user.number_of_posts,
    }


def _post(post_id):
    post = InstagramPost(post_id)
    return {
        "Post Id": post.post_id,
        "Author": post.author,
        "Likes": post.number_of_likes,
        "Comments": post.number_of_comments,
        "Date": post.upload_date,
        "Caption": post.caption,
    }


def _tag(tag):
    t = InstagramHashTag(tag)
    return {"Hashtag": "#" + t.tagname, "Posts": t.number_of_posts}


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description=__description__)

    parser.add_argument(
        "-u", "--user", required=False, help="Instagram Username", type=str
    )
    parser.add_argument(
        "-p", "--post", required=False, help="Instagram Post ID", type=str
    )
    parser.add_argument(
        "-t", "--tag", required=False, help="Instagram Hashtag name", type=str
    )
    parser.add_argument(
        "-v", "--version", help="Version of the Package", action="store_true"
    )

    args = parser.parse_args()

    username = args.user
    tag_name = args.tag
    post = args.post
    version = args.version

    if version or len(sys.argv) == 1:
        ver = sys.version_info
        _pprint(
            {
                "Python": f"{ver.major}.{ver.minor}.{ver.micro}",
                f"{__package__}": f"{__version__}",
            }
        )

    if username:
        _pprint(_user(username))
    elif tag_name:
        _pprint(_tag(tag_name))
    elif post:
        _pprint(_post(post))



"""
    instagramy.InstagramHashtag
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module scrape data
    of given Instagram Hashtag.

    Usage Example
    -------------
    ::

        >>> from instagramy.InstagramHashtag import InstagramHashtag

        >>> tag = InstagramHashtag('python')
        >>> tag.number_of_posts
        >>> tag.top_posts

"""

from .core.parser import Parser
from .core.parser import Viewer
from .core.parser import TagParser
from .core.exceptions import HashTagNotFound
from .core.exceptions import RedirectionError
from .core.exceptions import HTTPError
from .core.cache import Cache
from .core.requests import get


class InstagramHashTag(TagParser):
    r"""
    Scrapes instagram hashtag information

    :param tag: Name of the Instagram Hashtag
    :param sessionid (optional): Session id of Instagram which is in browser cookies
    :param from_cache (optional): Get data from the cache of instagramy not from instagram

    >>> hashtag = InstagramHashTag("python")
    >>> hashtag.number_of_posts
    3119668
    >>> instagram_user.posts_display_urls
    """

    def __init__(self, tag: str, sessionid=None, from_cache=False):
        self.url = f"https://www.instagram.com/explore/tags/{tag}/"
        self.sessionid = sessionid
        cache = Cache("tag")
        if from_cache:
            if cache.is_exists(tag):
                self.tag_data = cache.read_cache(tag)
            else:
                data = self.get_json()
                cache.make_cache(
                    tag, data["entry_data"]["TagPage"][0]["graphql"]["hashtag"]
                )
                self.tag_data = data["entry_data"]["TagPage"][0]["graphql"]["hashtag"]
        else:
            data = self.get_json()
            cache.make_cache(
                tag, data["entry_data"]["TagPage"][0]["graphql"]["hashtag"]
            )
            try:
                self.tag_data = data["entry_data"]["TagPage"][0]["graphql"]["hashtag"]
            except KeyError:
                raise RedirectionError
        if sessionid:
            try:
                self.viewer = Viewer(data=data["config"]["viewer"])
            except UnboundLocalError:
                self.viewer = None
        else:
            self.viewer = None

    def get_json(self) -> dict:
        """ Get Hashtag information from Instagram """

        try:
            html = get(self.url, sessionid=self.sessionid)
        except HTTPError:
            raise HashTagNotFound(self.url.split("/")[-2])
        parser = Parser()
        parser.feed(html)
        return parser.Data

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.tagname}')"

    def __str__(self) -> str:
        return f"{'#' + self.tagname} has {self.number_of_posts} posts"



""" Caches Management """

import os
import json
import shutil

cache_dir = ".instagramy_cache"


class Cache:

    """ Class for caches Management """

    def __init__(self, key: str):
        self.key = key
        if not os.path.isdir(cache_dir):
            os.mkdir(cache_dir)
        with open(cache_dir + "/CACHEDIR.TAG", "w") as file:
            file.write(
                "# This file is a cache directory tag created by instagramy." + "\n"
            )

    def is_exists(self, name: str) -> bool:
        return os.path.isfile(cache_dir + f"/{name}_{self.key}" + ".json")

    def make_cache(self, name: str, data: dict):
        with open(cache_dir + f"/{name}_{self.key}" + ".json", "w") as file:
            json.dump(data, file)

    def read_cache(self, name: str) -> dict:
        with open(cache_dir + f"/{name}_{self.key}" + ".json", "r") as file:
            return json.load(file)


def list_caches() -> None:
    """ List of all Cache files created by instagramy in current dir """

    return os.listdir(cache_dir)


def clear_caches() -> None:
    """ Clear all Caches created by instagramy in current dir """

    return shutil.rmtree(cache_dir, ignore_errors=True)



"""
    instagramy.InstagramUser
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module scrape data
    of given Instagram User.

    Usage Example
    -------------
    ::

        >>> from instagramy.InstagramUser import InstagramUser

        >>> user = InstagramUser('github')
        >>> user.is_verified
        >>> user.number_of_followers
        >>> user.biography

"""

from .core.parser import Parser
from .core.parser import Viewer
from .core.parser import UserParser
from .core.exceptions import UsernameNotFound
from .core.exceptions import RedirectionError
from .core.exceptions import HTTPError
from .core.cache import Cache
from .core.requests import get


class InstagramUser(UserParser):
    r"""
    Scrapes instagram user information.

    :param username: Username of the Instagram user
    :param sessionid (optional): Session id of Instagram which is in browser cookies
    :param from_cache (optional): Get data from the cache of instagramy not from instagram

    >>> instagram_user = InstagramUser("github")
    >>> instagram_user.is_verified
    True
    >>> instagram_user.biography
    'Built for developers.'
    """

    def __init__(self, username: str, sessionid=None, from_cache=False):

        self.url = f"https://www.instagram.com/{username}/"
        self.sessionid = sessionid
        cache = Cache("user")
        if from_cache:
            if cache.is_exists(username):
                self.user_data = cache.read_cache(username)
            else:
                data = self.get_json()
                cache.make_cache(
                    username, data["entry_data"]["ProfilePage"][0]["graphql"]["user"]
                )
                self.user_data = data["entry_data"]["ProfilePage"][0]["graphql"]["user"]
        else:
            data = self.get_json()
            cache.make_cache(
                username, data["entry_data"]["ProfilePage"][0]["graphql"]["user"]
            )
            try:
                self.user_data = data["entry_data"]["ProfilePage"][0]["graphql"]["user"]
            except KeyError:
                raise RedirectionError
        if sessionid:
            try:
                self.viewer = Viewer(data=data["config"]["viewer"])
            except UnboundLocalError:
                self.viewer = None
        else:
            self.viewer = None

    def get_json(self) -> dict:
        """ Get user information from Instagram """

        try:
            html = get(self.url, sessionid=self.sessionid)
        except HTTPError:
            raise UsernameNotFound(self.url.split("/")[-2])

        parser = Parser()
        parser.feed(html)
        return parser.Data

    def __str__(self) -> str:
        return f"{self.fullname} ({self.username}) -> {self.biography}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.username}')"



"""
    instagramy.InstagramPost
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module scrape Instagram Post data
    for given Instagram Post id.

    Usage Example
    -------------
    ::

        >>> from instagramy.InstagramHashtag import InstagramPost

        >>> post = InstagramPost('CGeYX2OA61s')
        >>> post.author
        >>> post.number_of_likes
        >>> post.number_of_comments

"""

from .core.parser import Viewer
from .core.parser import Parser
from .core.parser import PostParser
from .core.exceptions import PostIdNotFound
from .core.exceptions import RedirectionError
from .core.exceptions import HTTPError
from .core.cache import Cache
from .core.requests import get


class InstagramPost(PostParser):
    r"""
    Scrape the post information
    `https://www.instagram.com/p/<post_id>/`
    `https://www.instagram.com/p/CGeYX2OA61s/`

    :param post_id: Id of the Instagram post  (From url of the post)
    :param sessionid (optional): Session id of Instagram which is in browser cookies
    :param from_cache (optional): Get data from the cache of instagramy not from instagram

    >>> post = InstagramPost("CGeYX2OA61s")
    >>> post.author
    '@virat.kohli'
    >>> post.number_of_likes
    2203830
    >>> post.number_of_comments
    4629
    """

    def __init__(self, post_id: str, sessionid=None, from_cache=False):
        self.post_id = post_id
        self.url = f"https://www.instagram.com/p/{post_id}/"
        self.sessionid = sessionid
        cache = Cache("post")
        if from_cache:
            if cache.is_exists(post_id):
                self.post_data = cache.read_cache(post_id)
            else:
                data = self.get_json()
                cache.make_cache(
                    post_id,
                    data["entry_data"]["PostPage"][0]["graphql"]["shortcode_media"],
                )
                self.post_data = data["entry_data"]["PostPage"][0]["graphql"]["shortcode_media"]
        else:
            data = self.get_json()
            cache.make_cache(
                post_id, data["entry_data"]["PostPage"][0]["graphql"]["shortcode_media"]
            )
            try:
                self.post_data = data["entry_data"]["PostPage"][0]["graphql"][
                    "shortcode_media"
                ]
            except KeyError:
                raise RedirectionError
        if sessionid:
            try:
                self.viewer = Viewer(data=data["config"]["viewer"])
            except UnboundLocalError:
                self.viewer = None
        else:
            self.viewer = None

    def get_json(self) -> dict:
        """ Get post information from Instagram """

        try:
            html = get(self.url, sessionid=self.sessionid)
        except HTTPError:
            raise PostIdNotFound(self.post_id)
        parser = Parser()
        parser.feed(html)
        info = parser.Data
        return info

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.post_id}')"

    def __str__(self) -> str:
        return f"Post ({self.post_id}) posted by {self.author} with {self.number_of_likes} likes"



""" Plugins for Analyze Instagram """

from instagramy.InstagramUser import InstagramUser
from instagramy.InstagramHashTag import InstagramHashTag

__all__ = ["analyze_users_popularity", "analyze_hashtags", "analyze_user_recent_posts"]


def analyze_users_popularity(usernames: list, sessionid: str) -> dict:
    """ Functions return the required data to Analze Instagram users """

    followers = []
    following = []
    posts = []
    for username in usernames:
        user = InstagramUser(username, sessionid)
        followers.append(user.number_of_followers)
        following.append(user.number_of_followings)
        posts.append(user.number_of_posts)
    data = {
        "Usernames": usernames,
        "Followers": followers,
        "Following": following,
        "Posts": posts,
    }
    return data


def analyze_user_recent_posts(username: str, sessionid: str) -> dict:
    """ Functions return the required data to Analze Instagram user recent post """

    user = InstagramUser(username, sessionid)
    posts = user.posts
    urls = []
    likes = []
    comments = []
    for post in posts:
        urls.append(post["post_url"])
        likes.append(post["likes"])
        comments.append(post["comments"])

    return {"Posts": urls, "Likes": likes, "Comments": comments}


def analyze_hashtags(hashtags: list, sessionid: str) -> dict:
    """ Functions return the required data to Analze Instagram Hashtags """

    posts = []
    for hashtag in hashtags:
        tag = InstagramHashTag(hashtag, sessionid)
        posts.append(tag.number_of_posts)

    return {"Hashtag": hashtags, "Posts": posts}



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="instagramy",
    version="4.4",
    license='MIT',
    author="Yogeshwaran R",
    author_email="yogeshin247@gmail.com",
    description="Python Package for Instagram User, Posts, Hashtags and Locations data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yogeshwaran01/instagramy/",
    packages=setuptools.find_packages(),
    download_url="https://github.com/yogeshwaran01/instagramy/archive/master.zip",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



""" Plugins for Download image from Instagram """

import os
import urllib.request

from instagramy.InstagramUser import InstagramUser
from instagramy.InstagramPost import InstagramPost
from instagramy.InstagramHashTag import InstagramHashTag

__all__ = ["download_profile_pic", "download_post", "download_hashtags_posts"]


def download_profile_pic(username: str, sessionid=None, filepath=None) -> tuple:
    """ Download Instagram User Profile Picture """

    user = InstagramUser(username, sessionid)
    if filepath is None:
        filepath = f"{username}.jpg"
    pic_url = user.profile_picture_url
    return urllib.request.urlretrieve(pic_url, filename=filepath)


def download_post(id: str, sessionid=None, filepath=None) -> tuple:
    """ Download Instagram Post """

    post = InstagramPost(id, sessionid)
    if filepath is None:
        filepath = f"{id}.mp4"
    post_url = post.post_source
    return urllib.request.urlretrieve(post_url, filename=filepath)


def download_hashtags_posts(tag: str, sessionid: str, count=1):
    """
    Download posts of particualar Hashtag
    - It create the directory with name of give tagname
    - Download given count of posts with name of post id
    """

    if count > 65:
        raise Exception("Count must be less than 65")

    tag = InstagramHashTag(tag, sessionid)
    posts_ids = [post.shortcode for post in tag.top_posts][:count]
    os.mkdir(tag.tagname)
    for posts_id in posts_ids:
        post = InstagramPost(posts_id, sessionid)
        post_link = post.post_source
        if post.type_of_post == "GraphVideo":
            urllib.request.urlretrieve(
                post_link, filename=f"{tag.tagname}/{posts_id}.mp4"
            )
        else:
            urllib.request.urlretrieve(
                post_link, filename=f"{tag.tagname}/{posts_id}.jpg"
            )

    return True



""" Exception classes raised by Instagramy """

from urllib.error import HTTPError


class BaseException(HTTPError):
    """ Base Exception for instagramy """

    def __init__(self, name: str):
        self.name = name


class UsernameNotFound(BaseException):
    """ Raise if Username not found in Instagram """

    def __str__(self):
        return f"InstagramUser('{self.name}') not Found"


class HashTagNotFound(BaseException):
    """ Raise if Hashtag not found in Instagram """

    def __str__(self):
        return f"InstagramHashtag('{self.name}')"


class PostIdNotFound(BaseException):
    """ Raise if Post id not found in Instagram """

    def __str__(self):
        return f"InstargramPost('{self.name}')"


class LocationNotFound(BaseException):
    """ Raise if Location not found in Instagram """

    def __str__(self):
        return f"InstargramLocation('{self.name}')"


class RedirectionError(Exception):
    """ Raise if Instagram Redirects it to Login Page """

    def __str__(self):
        return "Instagram Redirects you to login page, \
Try After Sometime or Reboot your PC \
Provide the sessionid to Login"



""" Wrapper for urllib.request """

import random
from typing import Any
from urllib.request import Request, urlopen

from .user_agent import user_agents


def get(url: str, sessionid=None) -> Any:
    """
    Function send the HTTP requests to Instagram and
    Login into Instagram with session id
    and return the Html Content
    """

    request = Request(
        url=url, headers={"User-Agent": f"user-agent: {random.choice(user_agents)}"}
    )
    if sessionid:
        request.add_header("Cookie", f"sessionid={sessionid}")
    with urlopen(request) as response:
        html = response.read()

    return html.decode("utf-8")
